package debugger;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * Copyright (c) Ontos AG (http://www.ontosearch.com). This class is part of
 * JAPE Debugger component for GATE (Copyright (c) "The University of Sheffield"
 * see http://gate.ac.uk/) <br>
 * Contains helper static methods to access private class fields and methods.
 * 
 * @author Andrey Shafirin
 */

public class ClassRipper {
	/** Assumed as debug level between 0 (lowest or no debug) and 5 (highest) */
	private static int DebugLevel = 4;

	private ClassRipper() {
	}

	/**
	 * Obtains field from given class instance regardless either this field
	 * private or not
	 * 
	 * @param classInstance
	 *          class instance which field to take
	 * @param fieldName
	 *          field name
	 */
	private static Field getField(Object classInstance, String fieldName) {
		if (DebugLevel >= 5) {
			System.out.print("DEBUG [" + ClassRipper.class.getName() + "]:");
			System.out.print(" getField: ");
			System.out.print(" classInstance = [" + classInstance.getClass().getName() + "]");
			System.out.print(" fieldName = [" + fieldName + "]");
			System.out.print("\n");
		}
		Field field = null;
		Class cls = classInstance.getClass();
		while (!cls.getName().equals("java.lang.Object")) {
			try {
				field = cls.getDeclaredField(fieldName);
				field.setAccessible(true);
				return field;
			} catch (NoSuchFieldException e) {
				// will try to search in superclass
				// cls = classInstance.getClass().getSuperclass();
				cls = cls.getSuperclass();
			} catch (SecurityException e) {
				e.printStackTrace();
			}
		}
		return null;
	}

	/**
	 * Obtains method from the given class instance regardless either this field
	 * private or not.
	 * 
	 * @param classInstance
	 *          class instance which method to take
	 * @param methodName
	 *          method name
	 */
	private static Method getMethod(Object classInstance, String methodName, Class[] args) {
		if (DebugLevel >= 5) {
			System.out.print("DEBUG [" + ClassRipper.class.getName() + "]:");
			System.out.print(" getMethod: ");
			System.out.print(" classInstance = [" + classInstance.getClass().getName() + "]");
			System.out.print(" methodName = [" + methodName + "]");
			System.out.print("\n");
		}
		Class cls = classInstance.getClass();
		while (!cls.getName().equals("java.lang.Object")) {
			try {
				Method[] methods = cls.getDeclaredMethods();
				lbl_methods: for (int i = 0; i < methods.length; i++) {
					if (!methods[i].getName().equals(methodName)) continue;
					Class[] c = methods[i].getParameterTypes();
					if (c.length != args.length) continue;
					for (int j = 0; j < c.length; j++) {
						if (args[j] != null && !c[j].isAssignableFrom(args[j])) continue lbl_methods;
					}
					methods[i].setAccessible(true);
					return methods[i];
				}
				// will try to search in superclass
				cls = cls.getSuperclass();
			} catch (SecurityException e) {
				e.printStackTrace();
			}
		}
		return null;
	}

	/**
	 * Obtains field value from given class instance regardless either this field
	 * private or not
	 * 
	 * @param classInstance
	 *          class instance which value to extract
	 * @param fieldName
	 *          field name
	 * @return fileld value
	 */
	public static Object getFieldValue(Object classInstance, String fieldName) throws IllegalAccessException {
		Field field = getField(classInstance, fieldName);
		Object obj = field.get(classInstance);
		field.setAccessible(false);
		return obj;
	}

	public static boolean getBoolean(Object classInstance, String fieldName) throws IllegalAccessException {
		Field field = getField(classInstance, fieldName);
		boolean b = field.getBoolean(classInstance);
		field.setAccessible(false);
		return b;
	}

	public static byte getByte(Object classInstance, String fieldName) throws IllegalAccessException {
		Field field = getField(classInstance, fieldName);
		byte b = field.getByte(classInstance);
		field.setAccessible(false);
		return b;
	}

	public static char getChar(Object classInstance, String fieldName) throws IllegalAccessException {
		Field field = getField(classInstance, fieldName);
		char c = field.getChar(classInstance);
		field.setAccessible(false);
		return c;
	}

	public static int getInt(Object classInstance, String fieldName) throws IllegalAccessException {
		Field field = getField(classInstance, fieldName);
		int i = field.getInt(classInstance);
		field.setAccessible(false);
		return i;
	}

	public static float getFloat(Object classInstance, String fieldName) throws IllegalAccessException {
		Field field = getField(classInstance, fieldName);
		float f = field.getFloat(classInstance);
		field.setAccessible(false);
		return f;
	}

	public static long getLong(Object classInstance, String fieldName) throws IllegalAccessException {
		Field field = getField(classInstance, fieldName);
		long l = field.getLong(classInstance);
		field.setAccessible(false);
		return l;
	}

	public static double getDouble(Object classInstance, String fieldName) throws IllegalAccessException {
		Field field = getField(classInstance, fieldName);
		double d = field.getDouble(classInstance);
		field.setAccessible(false);
		return d;
	}

	public static Object invokeMethod(Object classInstance, String methodName, Object[] args)
			throws InvocationTargetException, IllegalAccessException {
		Class[] c = new Class[args.length];
		for (int i = 0; i < args.length; i++) {
			c[i] = (args[i] == null) ? null : args[i].getClass();
		}
		Method method = getMethod(classInstance, methodName, c);
		Object result = method.invoke(classInstance, args);
		method.setAccessible(false);
		return result;
	}
}
